// CForth LowPower project ***************************************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforthLP.c","cforthLP.h",
// and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{

START:
    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //Power Interface clock enable
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;  // Enable Port A clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;  // Enable Port B clock

	GPIOA->MODER = 0x00000000;      //primarily all input
	GPIOB->MODER = 0x00000000;      //primarily all input
	GPIOA->PUPDR = 0xAA8AAAAA;      //primarily all inputs pulldown,
									//except PA10
	GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown

//all default = 921kHz
#ifdef CLOCK_14M
	gfSpeed = 16;
#else
	gfSpeed = 4;
#endif
	SetSysClock();

//pin config repeated specifically:
#ifdef MODUL_24
	GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPDR11);	//pull-down off, wasts current
	GPIOA->MODER &= ~(GPIO_MODER_MODE11);
	GPIOA->MODER |= GPIO_MODER_MODE11_0 ;  	// Set PortA11 (LED) as output
	GPIOA->BSRR =  GPIO_BSRR_BS_11;        	//PA11 (LED) ON
#else
	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPDR0);	//pull-down off, wasts current
	GPIOB->MODER &= ~(GPIO_MODER_MODE0);
	GPIOB->MODER |= GPIO_MODER_MODE0_0 ;  	// Set PortB0 (LED) as output
	GPIOB->BSRR =  GPIO_BSRR_BS_0;        	//PB0 (LED) ON
#endif
	gSYSCOUNT = 50;
	gLEDCOUNT = 0;
	gDELAYCOUNT = 0;
	gTIMEOUT = 0;
	gBaud = DEFAULT_BAUD;
	gBaudSet = 0;
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
	gLedBusy = 3;
	gfRandom = 314159265;		//seed for xorshift32 algorithm

    SysTick_Config(SystemCoreClock / (1000 * gfSpeed));	
							//for any SystemCoreClock
							//SystemCoreClock is derived from SYSCLOCK=HSE
	wait5us();

//init I/O state  *********************
	popFlashPage23();           //loads only gBaud, gfNumBase and gProject
								//if virgin--> emptyForth();
	SysTick_Config(SystemCoreClock / (gfSpeed * 1000));	

//init communication  *********************
	initUsart1();
	msDelay(5);

	sendString(gTypeStr);
	sendTB(HARDWARE_VERSION + 0x30);
	sendTB('.');
	sendNS32(REVNUM);
	sendString(gCopyStr);

	popFlashProject();					//now load complete project
										//if virgin or gProject==0 emptyForth();

	while(1) {
		quit();
		if (query() == 0) break;
		if (gfTibMax == 0) continue;
		eval();
	}
	msDelay(100);
	goto START;       //force cold reset
}	
	

//##########################################################################

// Misc Low Level Routines
//************************
void SetSysClock(void)  // *************************************************
//assumes "Vcore range 1", HSE clock 3.6864 MHz or 14.7456MHz
//SystemClock=AHB=default 921.6kHz, may be changed to 3.6864 or 14.7456MHz
//APB1+2 = AHB clock
//adopted from "system_stm32f0xx.c" and modified according to STM32G431 FRefMan
//Reset SW[1:0], HPRE[3:0], PPRE1[2:0], PPRE2[2:0], MCOSEL[2:0] and MCOPRE[2:0]
// **************************************************************************
{
uint32_t StartUpCounter = 0, HSEStatus = 0;

//RCC_CR reset value = 0x00000063 (are all reserved entries)
//RCC_CFGR reset value = 0x00000005 (means HSI clock is on)

#define HSE_STARTUP_TIMEOUT   0x50000 //Time out for HSE start up

// SYSCLK, HCLK, PCLK configuration ----------------------------------------
// Enable HSE
    RCC->CR |= RCC_CR_HSEON;

// Wait till HSE is ready or Timeout is reached
    do  {
      HSEStatus = RCC->CR & RCC_CR_HSERDY;
      StartUpCounter++;
     } while((HSEStatus == 0) && (StartUpCounter  < HSE_STARTUP_TIMEOUT));

    if (HSEStatus) {
// Enable Prefetch Buffer and set Flash Latency
	FLASH->ACR =
        FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN; // zero waitstate

	RCC->CFGR &= ~(RCC_CFGR_HPRE);					  //clear all HPRE options
	if (gfSpeed == 16)  RCC->CFGR |= RCC_CFGR_HPRE_DIV16;//CLOCK_14M low speed
	else if (gfSpeed == 4)  RCC->CFGR |= RCC_CFGR_HPRE_DIV4; //3.68M low speed
    else  RCC->CFGR |= RCC_CFGR_HPRE_DIV1;				//any clock high speed
 // PCLK2 = HCLK
	RCC->CFGR |= RCC_CFGR_PPRE2_DIV1;
 //PCLK1 = HCLK
	RCC->CFGR |= RCC_CFGR_PPRE1_DIV1;

#ifdef CLOCK_14M			//14.7456MHz crystal
	SystemCoreClock    = 14745600;
#else						//only alternative 3.6864MHz crystal
	SystemCoreClock    = 3686400;
#endif

// DISable the main PLL explicitly
	RCC->CR &= ~RCC_CR_PLLON;
	RCC->CCIPR = (2 << 28) | 1;			//SYSCLK selected for ADC12 and USART1
// Select HSE as system clock source
     RCC->CFGR |= RCC_CFGR_SW_HSE;
     RCC->CFGR &= ~RCC_CFGR_SW_HSI;		//disable HSI
// Wait till HSE is used as system clock source
	while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_HSE) ;

	RCC->AHB1ENR |= RCC_AHB1ENR_FLASHEN;//necessary for Flash read/write action
    }
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
	gDELAYCOUNT = count;
	while(gDELAYCOUNT)  {
		servSystem();		//unconditionally in this case
	}
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// **************************************************************************
{
uint32_t countdown;

	countdown = 2;
    while (countdown)   {
        countdown-- ;
    }
}


void servSystem(void)   // ************************************************
//system relevant, but not extremely time critical services
//are handled quasi in the background
//but watchdog resets when it hangs somewhere else
//slowed down by SysTick to 20 times/second
// ****************************************************************************
{
	if (gSYSCOUNT)  return;		//countdown in SysTick_Handler()

//else if (gSYSCOUNT==0) restart 20Hz service interval
	gSYSCOUNT = 50;			

//LED control
#ifdef MODUL_24
	if (gLedBusy > 1)  {
		if (gLEDCOUNT) gLEDCOUNT--;
		if  (gLEDCOUNT==0) 	GPIOA->BSRR = GPIO_BSRR_BS_11;
											// Set bit PA11 (LED) ON again
	}
//as precaution repeated here:
	else if (gLedBusy ==0)  GPIOA->BSRR = GPIO_BSRR_BR_11;//LED always OFF
	else  GPIOA->BSRR = GPIO_BSRR_BS_11;		//gLedBusy=1:LED always ON
#else
	if (gLedBusy > 1)  {
		if (gLEDCOUNT) gLEDCOUNT--;
		if  (gLEDCOUNT==0) 	GPIOB->BSRR = GPIO_BSRR_BS_0;
											// Set bit PB0 (LED) ON again
	}
//as precaution repeated here:
	else if (gLedBusy == 0)  GPIOB->BSRR = GPIO_BSRR_BR_0;//LED always OFF
	else  GPIOB->BSRR = GPIO_BSRR_BS_0;		//gLedBusy=1: LED always ON
#endif
}


void popFlashPage23(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test1, test2;

	emptyForth();			//primarily init defaults - implies gProject = 0;

	flashPtr = (uint32_t *)PAGE23;

	test1 = ~(*flashPtr);//virgin test can't be=0 if project was saved in Flash
	flashPtr++;
	test2 = ~(*flashPtr);	//cannot be==0 if any project was saved in Flash
	flashPtr++;

	if ((test1 | test2)== 0)  return;	//defaults already init

	gBaud = ~(*flashPtr);			//baud rate is project independent
	if (gBaud == 0) gBaud = DEFAULT_BAUD;
	flashPtr++;
	test1 = ~(*flashPtr);//virgin test can't be=0 if project was saved in Flash
	if (test1) {			//Flash is not virgin, Project seems to be stored
		gfNumBase =  test1  & 0x000000FF;
		if (gfNumBase == 0) gfNumBase = DEFAULT_NUMBASE;
		gfTempNumBase = gfNumBase;
		gProject =  test1 >> 8;		//if==0, "empty project" will be used
	}
}


uint32_t pushFlashPage23(void)   //******************************************
//save project independent system parameters permanently
// **************************************************************************
{
uint32_t *flashPtr, *sramPtr;
uint32_t k, word;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
    else return 0;

//erase Page23
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((23<<3) | FLASH_CR_EOPIE | FLASH_CR_ERRIE | FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write mirror into Page23:
//this microcontroller allows exclusively doubleword-wide write in flash
//flash least word is first word at readout
//only write gBaud,gfNumBase and gProject,
//i.e.project independent parameters into page23

    flashPtr = (uint32_t *)PAGE23;
	FLASH->CR |= FLASH_CR_PG;

	*flashPtr = 0;	//for flash virgin test
	flashPtr++;
	*flashPtr = 0;	//for flash virgin test
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;

	*flashPtr = ~gBaud;
	flashPtr++;
	*flashPtr = ~((gProject << 8) | gfNumBase);
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;

//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return  1;
}


void popFlashProject(void)   // *********************************************
//first read some project specific globals
//next read 2 pages to gfUserHeader[], then read 6 pages to gfTCode
// **************************************************************************
{
uint32_t k, m, n, word, test;
uint32_t * flashPtr;

//emptyForth() already updated in popFlashPage23() (default system parameters)

	if (gProject)  {
		flashPtr = (uint32_t *)(PAGE24 + 0x4000 * (gProject - 1));
//first 32 bytes with system & peripheral config parameters
		test =  ~(*flashPtr);	//cannot be==0 if any project was saved in Flash
		flashPtr++;
		if (test == 0) {	//project Flash is virgin
			emptyForth();	//implies gProject=0 -> empty project
		}
		else  {
			word = ~(*flashPtr);
			flashPtr++;
//bit values of "word":
//bits 16..23: gfSpeed;
//bits 8-15: gfNumBase;
//bits 0..7: gLedBusy
			gfSpeed = (word >> 16) & 0x000000FF;
#ifdef CLOCK_14M
			if ((gfSpeed !=1)&&(gfSpeed !=16)) gfSpeed = 16;//default 921.6kHz
			RCC->CFGR &= ~(RCC_CFGR_HPRE_Msk);		//clear all HPRE options
			if (gfSpeed == 16)  RCC->CFGR |= RCC_CFGR_HPRE_DIV16;
			else  RCC->CFGR |= RCC_CFGR_HPRE_DIV1;
#else
			if ((gfSpeed != 1)&&(gfSpeed != 4)) gfSpeed = 4;//default 921.6kHz
			RCC->CFGR &= ~(RCC_CFGR_HPRE_Msk);		//clear all HPRE options
			if (gfSpeed == 4)  RCC->CFGR |= RCC_CFGR_HPRE_DIV4;
			else  RCC->CFGR |= RCC_CFGR_HPRE_DIV1;
#endif
			gfNumBase = (word >> 8) & 0x000000FF;
			gLedBusy = word & 0x000000FF;
			if (gLedBusy > 3) gLedBusy = 2;

			if ((gfNumBase != 10)&&(gfNumBase != 16)) gfNumBase = 10;
															//default DECIMAL
			gfTempNumBase = gfNumBase;
			gfAuto = ~(*flashPtr);
			flashPtr ++;

			word = ~(*flashPtr);
			flashPtr ++;
			gfSpiMode = (word >> 24) & 0x00000003;
			gfSpiBits = (word >> 16) & 0x00000003F;
			gfSpiClock = word & 0x00000FFFF;

			gfBaseFreq12 = ~(*flashPtr);
			flashPtr ++;
#ifdef CLOCK_14M
			gfFreq12 = 16 * gfBaseFreq12 / gfSpeed;
#else
			gfFreq12 = 4 * gfBaseFreq12 / gfSpeed;
#endif
			k = ~(*flashPtr);
			gfPWM1 = (int32_t)k;
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM2 = (int32_t)k;
			flashPtr ++;
			if (gfFreq12) {
				subFREQ12();
				subPWM1();
				subPWM2();
			}
			k = ~(*flashPtr);
			flashPtr ++;
			gfPWM3 = (int32_t)(k & 0x0000FFFF);
			if ((gfPWM3 == 0) || (gfPWM3 == 0x0000FFFF)) gfPWM3 = -1;
			gfBaseFreq3 = ((k >> 16) & 0x0000FFFF);
#ifdef CLOCK_14M
			gfFreq3 = 16 * gfBaseFreq3 / gfSpeed;
#else
			gfFreq3 = 4 * gfBaseFreq3 / gfSpeed;
#endif
			if (gfFreq3) {
				subFREQ3();
				subPWM3();
			}
//Flash readput is possible over Flash pages
			for (k = 0;k < MAXUSER/2;k++)  {  
										//reads 84*24 words=2016bytes +32=2048
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
			flashPtr = (uint32_t *)(PAGE25 + 0x4000 * (gProject - 1));
			for (k = (MAXUSER/2) +1;k < MAXUSER;k++) {  //reads 2044 bytes!
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
	
			flashPtr = (uint32_t *)(PAGE26 + 0x4000 * (gProject - 1));
			for (k = 0;k < MAXCODE;k++)  {	//reads 6 pages each 2048 bytes!
				gfTCode[k] = ~(*flashPtr);
				flashPtr ++;
			}
		}
	}

//from here if(gProject != 0) or virgin or not:
	wait5us();
	op_ABORT();							//reset project Forth system data
	if (gfAuto)  {
//emergency brake to avoid AUTO deadlock
		if ((GPIOA->IDR & 0x400) == 0) gfAuto = 0;	//if PA10 (RX input)is low
	}
	sendTB(CRR);
	op_QUEST();
	if (gfAuto) {
        gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
		interpretUserToken();
	}
}


uint32_t pushFlashProject(void)   //*****************************************
// **************************************************************************
{
uint32_t k, m, n, offset;
uint32_t * flashPtr;
uint32_t word[12];

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

	if ((gProject == 0) || (gProject > 5)) return 0;//precaution
													//don't save empty project
	
    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

    offset = (gProject - 1) * 8;	//project no.0 is empty, not treated here
									//project no.1 starts with page#24
//erase Page24 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((24+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//this microcontroller allows exclusively doubleword-wide write in flash
//first write virgin test, gfNumBase and gfAuto into Page24 etc

    flashPtr = (uint32_t *)(PAGE24 + 0x800 * offset);
    FLASH->CR |= FLASH_CR_PG;
//first 32 bytes with system & peripheral config parameters
    *flashPtr = 0x00000000; //virgin flash test
    flashPtr++;
    *flashPtr = ~((gfSpeed << 16) | (gfNumBase << 8) | gLedBusy );
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;

    *flashPtr = ~gfAuto;
    flashPtr++;
    *flashPtr = (~((gfSpiMode << 24) | (gfSpiBits << 16 ) |gfSpiClock));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfBaseFreq12;
	flashPtr++;
    *flashPtr = ~((uint32_t)gfPWM1);
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~((uint32_t)gfPWM2);
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfBaseFreq3 << 16) & 0xFFFF0000) | 
								((uint32_t)gfPWM3 & 0x0000FFFF));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

//write first half of gfUserHeader[] into Page24 + offset
    for (k = 0;k < MAXUSER/2; k+=2) {	//MAXUSER/2 = 84 ->2016bytes+32=2048
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;

        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;

        for (m = 0; m < 12; m+=2)  {
// this microcontroller allows exclusively doubleword-wide write in flash
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility w. empty SRAM
            flashPtr++;
//burn
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page25 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((25+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second half of gfUserHeader[] into Page25 + offset
    flashPtr = (uint32_t *)(PAGE25+0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = (MAXUSER/2);k < MAXUSER; k+=2) {     //MAXUSER/2 = 84
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;
        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;
        for (m = 0; m < 12; m+=2)  {
// this microcontroller allows exclusively doubleword-wide write in flash
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility w.empty SRAM
            flashPtr++;
//burn
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page26 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((26+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write first part of gTCode into Page26 + offset
    flashPtr = (uint32_t *)(PAGE26 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0; k < 0x200; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page27 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((27+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second part of gTCode into Page20 + offset

    flashPtr = (uint32_t *)(PAGE27 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x200; k < 0x400; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page28 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((28+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    FLASH->CR &= ~FLASH_CR_PER;

//write third part of gTCode into Page28
    flashPtr = (uint32_t *)(PAGE28 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x400; k < 0x600; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page29 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((29+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write fourth part of gTCode into Page29 + offset
    flashPtr = (uint32_t *)(PAGE29 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x600; k < 0x800; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page30 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((30+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
     FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;
    FLASH->CR &= ~FLASH_CR_PER;

//write fifth part of gTCode into Page30 + offset
    flashPtr = (uint32_t *)(PAGE30 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x800; k < 0xA00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page31 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((31+offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write sixth part of gTCode into Page31 + offset
    flashPtr = (uint32_t *)(PAGE31 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0xA00; k < 0xC00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB(SPACE);

//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return 1;
}


//###########################################################################

//USART Support
//*************

void initUsart1(void) // ****************************************************
// USART1-TxD is PortB,6
// USART1-RxD is PortA,10
// **************************************************************************
{
uint16_t baudset;

    RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ;   // USART1 clock enable

//set PA10 in alternate function mode
    GPIOA->MODER &= ~GPIO_MODER_MODE10_0 ;
    GPIOA->MODER |= GPIO_MODER_MODE10_1;
//set PB6 in alternate function mode
    GPIOB->MODER &= ~GPIO_MODER_MODE6_0 ;
    GPIOB->MODER |= GPIO_MODER_MODE6_1;

//USART1 is AF7 on PortA
    GPIOA->AFR[1] &= 0xFFFFF0FF;        //PA10 is RX
    GPIOA->AFR[1] |= 0x00000700;
//and AF7 on PortB
    GPIOB->AFR[0] &= 0xF0FFFFFF;        //PB6 is TX
    GPIOB->AFR[0] |= 0x07000000;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

#ifdef CLOCK_14M
    switch(gBaud) {
    case 96:
        baudset = 1536;       //9600Baud: 1536 at 14.7456MHz SYSCLK
        break;
    case 38:
        baudset = 384;       //38400Baud: 384 at 14.7456MHz SYSCLK
        break;
    case 19:
        baudset = 768;       //19200Baud: 768 at 14.7456MHz SYSCLK
        break;
    case 57:
        baudset = 256;       //57600Baud: 256 at 14.7456MHz SYSCLK
        break;
    case 11:
    default:
        baudset = 128;       //115200Baud: 128 at 14.74564MHz SYSCLK
        break;
    }
#else
    switch(gBaud) {
    case 96:
        baudset = 384;       //9600Baud: 384 at 3.6864MHz SYSCLK
        break;
    case 38:
        baudset = 96;       //38400Baud: 96 at 3.6864MHz SYSCLK
        break;
    case 19:
        baudset = 192;       //19200Baud: 192 at 3.6864MHz SYSCLK
        break;
    case 57:
        baudset = 64;       //57600Baud: 64 at 3.6864MHz SYSCLK
        break;
	case 11:
    default:
        baudset = 32;       //115200Baud: 32 at 3.6864MHz SYSCLK
        break;
    }
#endif
    USART1->BRR = baudset;
	USART1->CR2 = USART_CR2_TXINV;	//1 Stopbit, TX level inverted
    USART1->CR3 = 0;				//default:no Handshake
    USART1->CR1 = USART_CR1_RXNEIE | USART_CR1_TE | USART_CR1_RE;
									//8Bit, no Parity, RX Int
    USART1->CR1 |= USART_CR1_UE;	//enable USART
    gBaudSet = 1;
    NVIC_EnableIRQ(USART1_IRQn);
}


void sendTB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
	servSystem();	//serve watchdog and LED
    while(!(USART1->ISR & USART_ISR_TXE));
    USART1->TDR = data;
	if (gLedBusy == 3)  {
		gLEDCOUNT = 4;					//ca. 300 milliseconds
#ifdef MODUL_24
		GPIOA->BSRR = GPIO_BSRR_BR_11; // clear bit PA11,LED OFF
#else
		GPIOB->BSRR = GPIO_BSRR_BR_0; // clear bit PB0,LED OFF
#endif
	}
}

void sendNibble(uint8_t data) // ********************************************
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;

        sendTB(asciiNibble);
}


void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // *********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    sendTB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendTB(0x20);
}

void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;

    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
//	if (data >= 10)  {			//modified 11Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(0x20);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(0x20);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}


uint32_t pullTermRxBuf(void) // ********************************************
// pull one byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();	//serve watchdog and LED

	if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
		recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
		if (gUSART_RXTAIL==gUSART_RXBUFTOP)  gUSART_RXTAIL=gUSART_RXBUF;

		if (gLedBusy > 1)  {
			gLEDCOUNT = 4;					//ca. 300 milliseconds
#ifdef MODUL_24
			GPIOA->BSRR = GPIO_BSRR_BR_11;	//clear bit PA11,LED OFF
#else
			GPIOB->BSRR = GPIO_BSRR_BR_0;	//clear bit PB0,LED OFF
#endif
		}
		return recv;
	}
	else return 0;
}

// ##########################################################################
// Interrupt Handlers


void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
	if (gSYSCOUNT > 0) gSYSCOUNT--;			//controls servSystem()
	if (gDELAYCOUNT > 0) gDELAYCOUNT--;
	if (gTIMEOUT > 0) gTIMEOUT--;			//to be used in any subroutine
	if (gfMSCOUNT > 0) gfMSCOUNT --;		//used by op_MS,op_DS
	if (gfTIXCOUNT > 0) gfTIXCOUNT --;		//used by op_TIX,op_WAIT
	if (gfSTEPCOUNT > 0) gfSTEPCOUNT--;		//stepper timing
	gfRandom ^= gfRandom << 13;				//xorshift32 algorithm forRANDOM
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;	
}


void USART1_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t test;

	if (USART1->ISR & USART_ISR_RXNE)  {
		test = USART1->RDR & 0x00FF;
		if (test == CTRL_B) gfTermInput = CTRL_B;
		if (test != CTRL_D)  {
			*gUSART_RXHEAD = (uint8_t)test;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
		else  op_DEBUG();
	}
}

//  ####### end of file #####################################################