// main.h Header file of the CForth LowPower project *************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "cforthLP.c", "cforthLP.h",
// and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "globaldefines.h"

// from cforth.h
extern uint32_t gfMSCOUNT, gfTIXCOUNT;
extern uint32_t gfTibMax;
extern uint32_t gfNumBase, gfTempNumBase, gfAuto, gfSpeed;
extern uint32_t gfBackProcess;
extern uint32_t gfTermInput;		//message from terminal iput to DEBUG();
extern uint32_t gfSpiBits, gfSpiMode, gfSpiClock;
extern uint32_t gfFreq12, gfBaseFreq12, gfFreq3, gfBaseFreq3;
extern int32_t gfPWM1, gfPWM2, gfPWM3;
extern int32_t gfSTEPCOUNT;
extern uint32_t gfCompiState;		//0=interpreting, 1=compiling
extern uint32_t gfErrors, gfParsed;
extern uint32_t gfRandom;
extern uint32_t *gfSP;				//Datastack Pointer
extern uint32_t *gfRP;				//Returnstack Pointer
extern uint32_t *gfIP;				//Instruction Pointer
extern uint32_t gfTCode[];

extern struct USERHEADER
{
uint8_t name[16];
uint32_t toCompile;
uint32_t attributes;
} gfUserHeader[MAXUSER];
//6 words = 24 bytes/struct * 168 entries = 1008 words = 4032 bytes)

extern void quit(void);
extern uint32_t query(void);
extern void eval(void);
extern uint32_t interpretUserToken(void);
extern uint32_t executeTCode(void);
extern void emptyForth(void);
extern void op_NOP(void);
extern void subFREQ12(void);
extern void subPWM1(void);
extern void subPWM2(void);
void subFREQ3(void);
void subPWM3(void);
extern void op_DEBUG(void);
extern void op_ABORT(void);
extern void op_QUEST(void);

// ****************************************

//buffer length defined in "globaldefines.h"
uint8_t gUSART_RXBUF[RXBUFLEN+4];
uint8_t *gUSART_RXHEAD=0,*gUSART_RXTAIL=0;
uint8_t *gUSART_RXBUFTOP= &(gUSART_RXBUF[RXBUFLEN]);

//system control
uint32_t gSYSCOUNT, gLEDCOUNT, gDELAYCOUNT, gTIMEOUT;
uint32_t gBaud, gBaudSet, gLedBusy, gProject;

#define PAGE23 (0x0800B800)	//starting from 46kB Flash: system globals 
//project 1 starting from 48kB Flash:
#define PAGE24 (0x0800C000)	//project parameters and UserToken struct proj.1
#define PAGE25 (0x0800C800)
#define PAGE26 (0x0800D000)	//project1 code
#define PAGE27 (0x0800D800)
#define PAGE28 (0x0800E000)
#define PAGE29 (0x0800E800)
#define PAGE30 (0x0800F000)
#define PAGE31 (0x0800F800)	//end code project1
//project 2 from 64kB: pages 32,33 param+UserToken, 34,35,36,37,38,39 code
//project 3:from 80kB: pages 40,41 param+UserToken, 42,43,44,45,46,47 code
//project 4:from 96kB: pages 48,49 param+UserToken, 50,51,52,53,54,55 code
//project 5:from 112kB:pages 56,57 param+UserToken, 58,59,60,61,62,63 code

#ifdef CLOCK_14M			//14.7456MHz crystal
const uint8_t gTypeStr [] = "CForth LowPower-921kHz/3.6/14.7MHz-for STM32G4xx v";
#else						//only alternative 3.6864MHz crystal
const uint8_t gTypeStr [] = "CForth LowPower -921kHz/3.6864MHz- for STM32G4xx v";
#endif
const uint8_t gCopyStr [] = " (c)2020-24 W.Schemmert";
const uint8_t gColdStartStr [] = "Cold Restart";

void SetSysClock(void);
void msDelay(uint32_t);
void wait5us(void);
void popFlashPage23(void);
uint32_t pushFlashPage23(void);
void popFlashProject();
uint32_t pushFlashProject(void);
void servSystem(void);
//USART
void initUsart1(void);
void sendTB(uint8_t);
void sendNibble(uint8_t);
void sendString(const uint8_t *);
void sendX8(uint32_t);
void sendX32(uint32_t);
void sendNX32(uint32_t);
void send0X32(uint32_t);
void sendLS32(uint32_t);
void sendS32(uint32_t);
void sendNS32(uint32_t);
uint32_t pullTermRxBuf(void);
uint8_t parseHexNibble(uint8_t);
uint32_t parseHexNum(uint8_t);
uint8_t parseDeciDigit(void);
//Interrupt handlers
//void SysTick_Handler(void);
//void USART1_IRQHandler(void);
